using Gst;

namespace Metronome {

    public class MetronomeApp : Granite.Application {

        public Gtk.Window mainwindow;

        // Controls
        Gtk.Paned root;
        Gtk.Grid content;
        Gtk.Box sidebar;
        Gtk.ListBox list_box;

        Gtk.Button start_button;

        Gtk.SpinButton tempo;
        Gtk.SpinButton beat;

        Gtk.Image led_green_right;
        Gtk.Image led_red_left;

        Gtk.Label tempo_value;
        Gtk.Label beat_value;

        Gtk.Scale volume_semibreve;
        Gtk.Scale volume_minim;
        Gtk.Scale volume_crotchet;
        Gtk.Scale volume_quaver;

        Gtk.Popover tempo_popover;
        Gtk.Popover beat_popover;

        Gst.Element sound_semibreve;
        Gst.Element sound_minim;
        Gst.Element sound_crotchet;
        Gst.Element sound_quaver;
        Gst.Element sound_beat;

        Metronome.Click click;

        Settings settings;

        string file_click_1 = "file://" + Constants.PKGDATADIR + "/sounds/click_1.ogg";
        string file_click_2 = "file://" + Constants.PKGDATADIR + "/sounds/click_2.ogg";
        string file_click_4 = "file://" + Constants.PKGDATADIR + "/sounds/click_4.ogg";
        string file_click_8 = "file://" + Constants.PKGDATADIR + "/sounds/click_8.ogg";
        string file_beat = "file://" + Constants.PKGDATADIR + "/sounds/beat.ogg";

        public static MetronomeApp _instance = null;

        public static MetronomeApp instance {
            get {
                if (_instance == null)
                    _instance = new MetronomeApp ();
                return _instance;
            }
        }

        construct {
            program_name = "Metronome";
            exec_name = "metronome";

            build_data_dir = Constants.DATADIR;
            build_pkg_data_dir = Constants.PKGDATADIR;
            build_release_name = Constants.RELEASE_NAME;
            build_version = Constants.VERSION;
            build_version_info = Constants.VERSION_INFO;

            app_years = "2015";
            app_icon = "metronome";
            app_launcher = "metronome.desktop";
            application_id = "net.launchpad.Metronome";

            main_url = "https://code.launchpad.net/metronome";
            bug_url = "https://bugs.launchpad.net/metronome";
            help_url = "https://code.launchpad.net/metronome";
            translate_url = "https://translations.launchpad.net/metronome";

            about_authors = {"Artem Anufrij <artem.anufrij@live.de>"};
            about_documenters = {"Artem Anufrij <artem.anufrij@live.de>"};
            about_artists = {"Sam Hewitt",
                                "Artem Anufrij <artem.anufrij@live.de>"};
            about_comments = "Development release, not all features implemented";
            about_translators = "";
            about_license_type = Gtk.License.GPL_3_0;
        }

        protected override void activate () {

            debug ("sound file: %s", file_click_1);
            debug ("sound file: %s", file_click_2);
            debug ("sound file: %s", file_click_4);
            debug ("sound file: %s", file_click_8);
            debug ("sound file: %s", file_beat);

            if (mainwindow != null) {
                mainwindow.present (); // present window if app is already running
                return;
            }

            settings = Settings.get_default ();

            click = new Metronome.Click ();
            click.current_tempo = settings.tempo;
            click.current_beat = settings.beat;

            click.state_changed.connect((state) => {
                if (state == click_state.running) {
                    start_button.label = _("Stop");
                    start_button.get_style_context ().remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION);
                    start_button.get_style_context ().add_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION);

                    led_green_right.visible = false;
                    led_red_left.visible = false;
                }
                else {
                    start_button.label = _("Start");
                    start_button.get_style_context ().remove_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION);
                    start_button.get_style_context ().add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION);

                    stop_sound ();
                }
            });

            click.tempo_changed.connect ((current_tempo) => {
                debug ("current_tempo %d", (int)current_tempo);
                tempo_value.label = "<b>" + current_tempo.to_string () + "</b>";
            });

            click.beat_changed.connect ((current_beat) => {
                debug ("current_beat %d", (int)current_beat);
                beat_value.label = "<b>" + current_beat.to_string () + "</b>";
            });

            click.click.connect ((interval, beat) => {
                do_click (interval, beat);
            });

            // SOUND
            sound_semibreve = ElementFactory.make ("playbin", "player");

            sound_minim = ElementFactory.make ("playbin", "player");

            sound_crotchet = ElementFactory.make ("playbin", "player");

            sound_quaver = ElementFactory.make ("playbin", "player");

            sound_beat = ElementFactory.make ("playbin", "player");

            build_ui ();
        }

        private void build_ui () {
            mainwindow = new Gtk.Window ();
            mainwindow.set_resizable (false);
            mainwindow.add_events(Gdk.EventMask.KEY_PRESS_MASK);

            root = new Gtk.Paned (Gtk.Orientation.HORIZONTAL);

            content = new Gtk.Grid ();
            content.width_request = 200;
            content.margin = 24;
            content.hexpand = true;
            content.row_spacing = 24;
            content.column_spacing = 6;

            // TITLE BAR
            build_titelbar ();

            // VOLUME sliders
            build_volume_sliders ();

            // SIDEBAR
            build_sidebar ();

            // START-STOP button
            start_button = new Gtk.Button.with_label (_("Start"));
            start_button.hexpand = true;
            start_button.get_style_context ().add_class ("h2");
            start_button.get_style_context ().add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION);
            content.attach (start_button, 1, 1, 2, 1);

            start_button.clicked.connect (() => {
                if (click.current_state == click_state.running)
                    click.stop ();
                else
                    click.start.begin ();
            });

            // LEDs
            led_green_right = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/led_green.png");
            content.attach (led_green_right, 3, 1, 1, 1);

            var led_gray_right = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/led_gray.png");
            content.attach (led_gray_right, 3, 1, 1, 1);

            led_red_left = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/led_red.png");
            content.attach (led_red_left, 0, 1, 1, 1);

            var led_gray_left = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/led_gray.png");
            content.attach (led_gray_left, 0, 1, 1, 1);

            mainwindow.add (root);
            mainwindow.show_all ();

            sidebar.visible = false;
            led_green_right.visible = false;
            led_red_left.visible = false;

            mainwindow.destroy.connect (() => {
                click.stop ();
                save_settings ();
                Gtk.main_quit ();
            });

            load_settings ();

            Gtk.main ();
        }

        private void build_titelbar () {
            var headerbar = new Gtk.HeaderBar ();
            headerbar.show_close_button = true;

            mainwindow.set_titlebar (headerbar);
            mainwindow.window_position = Gtk.WindowPosition.CENTER;

            var settings_buttons = new Granite.Widgets.ModeButton ();
            settings_buttons.margin_left = 24;
            settings_buttons.margin_right = 24;
            settings_buttons.mode_changed.connect(() => {
                if (settings_buttons.selected == 0) {
                    tempo_popover.show_all ();
                }
                else if (settings_buttons.selected == 1) {
                    beat_popover.show_all ();
                }
            });

            headerbar.set_custom_title (settings_buttons);

            // TEMPO button
            var tempo_button = new Gtk.Grid ();
            tempo_value = new Gtk.Label ("<b>%d</b>".printf (settings.tempo)) { use_markup = true };
            var tempo_bpm = new Gtk.Label ("<small>BPM</small>") { use_markup = true };

            tempo_button.attach (tempo_value, 0, 0, 1, 1);
            tempo_button.attach (tempo_bpm, 0, 1, 1, 1);

            // TEMPO input
            tempo = new Gtk.SpinButton.with_range (30, 240, 1);
            tempo.margin = 6;
            tempo.value_changed.connect (() => {
                set_new_tempo ((uint)tempo.value);
            });

            tempo.key_press_event.connect ((event) => {
                if (event.keyval == Gdk.Key.Return || event.keyval == Gdk.Key.KP_Enter) {
                    if (set_new_tempo ((uint)tempo.value)) {
                        tempo_popover.hide ();
                    }
                    return true;
                }
                else
                    click.stop ();
                return false;
            });

            // TEMPO popover
            tempo_popover = new Gtk.Popover (tempo_button);
            tempo_popover.position = Gtk.PositionType.BOTTOM;
            tempo_popover.add (tempo);

            // BEAT button
            var beat_button = new Gtk.Grid ();
            beat_value = new Gtk.Label ("<b>%d</b>".printf (settings.beat)) { use_markup = true };
            var beat_beat = new Gtk.Label ("<small>Beat</small>") { use_markup = true };

            beat_button.attach (beat_value, 0, 0, 1, 1);
            beat_button.attach (beat_beat, 0, 1, 1, 1);

            // BEAT input
            beat = new Gtk.SpinButton.with_range (1, 9, 1);
            beat.margin = 6;
            beat.value_changed.connect (() => {
                set_new_beat ((uint)beat.value);
            });

            beat.key_press_event.connect ((event) => {
                if (event.keyval == Gdk.Key.Return || event.keyval == Gdk.Key.KP_Enter) {
                    if (set_new_beat ((uint)beat.value)) {
                        beat_popover.hide ();
                    }
                    return true;
                }
                return false;
            });

            beat_popover = new Gtk.Popover (beat_button);
            beat_popover.position = Gtk.PositionType.BOTTOM;
            beat_popover.add (beat);

            settings_buttons.append (tempo_button);
            settings_buttons.append (beat_button);

            var sidebar_toogle = new Gtk.Button ();
            sidebar_toogle.image = new Gtk.Image.from_icon_name ("pane-show-symbolic-rtl", Gtk.IconSize.LARGE_TOOLBAR);
            sidebar_toogle.relief = Gtk.ReliefStyle.NONE;
            sidebar_toogle.clicked.connect (() => {
                if (sidebar.visible) {
                    sidebar.visible = false;
                    sidebar_toogle.image = new Gtk.Image.from_icon_name ("pane-show-symbolic-rtl", Gtk.IconSize.LARGE_TOOLBAR);
                } else {
                    sidebar.visible = true;
                    sidebar_toogle.image = new Gtk.Image.from_icon_name ("pane-hide-symbolic-rtl", Gtk.IconSize.LARGE_TOOLBAR);
                }
            });

            headerbar.pack_end (sidebar_toogle);

            /*var reset = new Gtk.Button ();
            reset.image = new Gtk.Image.from_icon_name ("document-revert-symbolic", Gtk.IconSize.LARGE_TOOLBAR);
            reset.tooltip_text = _("Revert all settings");
            reset.relief = Gtk.ReliefStyle.NONE;
            reset.clicked.connect (reset_settings);

            headerbar.pack_end (reset);*/
        }

        private void build_volume_sliders () {

            var volumes = new Gtk.Grid ();
            volumes.hexpand = true;
            volumes.row_spacing = 6;
            volumes.column_spacing = 6;
            volumes.height_request = 180;

            volume_semibreve = new Gtk.Scale.with_range (Gtk.Orientation.VERTICAL, 0, 1, 0.01);
            volume_semibreve.expand = true;
            volume_semibreve.inverted = true;
            volume_semibreve.draw_value = false;
            volume_semibreve.set_value (0.5);

            var icon_semibreve = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/symbol_semibreve.png");

            volume_minim = new Gtk.Scale.with_range (Gtk.Orientation.VERTICAL, 0, 1, 0.01);
            volume_minim.expand = true;
            volume_minim.inverted = true;
            volume_minim.draw_value = false;

            var icon_minim = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/symbol_minim.png");

            volume_crotchet = new Gtk.Scale.with_range (Gtk.Orientation.VERTICAL, 0, 1, 0.01);
            volume_crotchet.expand = true;
            volume_crotchet.inverted = true;
            volume_crotchet.draw_value = false;

            var icon_crotchet = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/symbol_crotchet.png");

            volume_quaver = new Gtk.Scale.with_range (Gtk.Orientation.VERTICAL, 0, 1, 0.01);
            volume_quaver.expand = true;
            volume_quaver.inverted = true;
            volume_quaver.draw_value = false;

            var icon_quaver = new Gtk.Image.from_file (Constants.PKGDATADIR + "/icons/symbol-quaver.png");

            volumes.attach (volume_semibreve, 0, 0, 1, 1);
            volumes.attach (icon_semibreve, 0, 1, 1, 1);
            volumes.attach (volume_minim, 1, 0, 1, 1);
            volumes.attach (icon_minim, 1, 1, 1, 1);
            volumes.attach (volume_crotchet, 2, 0, 1, 1);
            volumes.attach (icon_crotchet, 2, 1, 1, 1);
            volumes.attach (volume_quaver, 3, 0, 1, 1);
            volumes.attach (icon_quaver, 3, 1, 1, 1);

            content.attach (volumes, 0, 0, 4, 1);

            root.pack1 (content, true, false);
        }

        private void build_sidebar () {

            sidebar = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            sidebar.visible = false;
            sidebar.width_request = 140;

            var list_scroll = new Gtk.ScrolledWindow (null, null);
            list_scroll.vexpand = true;
            list_scroll.get_style_context().set_junction_sides(Gtk.JunctionSides.BOTTOM);

            // LIST BOX
            list_box = new Gtk.ListBox ();
            list_box.expand = true;
            list_box.row_activated.connect ((row) => {
                var custom_setting =  (((Metronome.SettingRow)row).setting);
                tempo.set_value (custom_setting.tempo);
                beat.set_value (custom_setting.beat);
            });

            list_scroll.add (list_box);

            var list_toolbar = new Gtk.Toolbar ();
            list_toolbar.expand = false;
            list_toolbar.set_style (Gtk.ToolbarStyle.ICONS);
            list_toolbar.set_icon_size (Gtk.IconSize.SMALL_TOOLBAR);
            list_toolbar.set_show_arrow (false);
            list_toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR);
            list_toolbar.get_style_context().set_junction_sides(Gtk.JunctionSides.TOP);



            // ADD Button
            var list_toolbar_add = new Gtk.ToolButton (null, null);
            list_toolbar_add.icon_name = "list-add-symbolic";
            list_toolbar.insert (list_toolbar_add, -1);

            // ADD Popover
            var popover_grid = new Gtk.Grid ();
            popover_grid.margin = 6;
            popover_grid.row_spacing = 6;
            popover_grid.column_spacing = 6;

            var new_setting_label = new Gtk.Label (_("Title:"));

            var new_setting_title = new Gtk.Entry ();
            new_setting_title.hexpand = true;
            new_setting_title.set_placeholder_text (_("Setting title"));

            var add_setting_popover = new Gtk.Popover (list_toolbar_add);
            add_setting_popover.position = Gtk.PositionType.TOP;
            add_setting_popover.add (popover_grid);

            var new_setting_add = new Gtk.Button.with_label (_("Add"));
            new_setting_add.clicked.connect (() => {
                if (new_setting_title.buffer.text != "") {
                    var new_custom_setting = new Metronome.Setting ();
                    new_custom_setting.title = new_setting_title.buffer.text;
                    new_custom_setting.tempo = click.current_tempo;
                    new_custom_setting.beat = click.current_beat;

                    add_custom_setting (new_custom_setting);
                    add_setting_popover.hide ();
                }
            });

            popover_grid.attach (new_setting_label, 0, 0, 1, 1);
            popover_grid.attach (new_setting_title, 1, 0, 1, 1);
            popover_grid.attach (new_setting_add, 0, 1, 2, 1);

            list_toolbar_add.clicked.connect (() => {
               new_setting_title.text = "";
               add_setting_popover.show_all ();
            });

            // REMOVE Button
            var list_toolbar_remove = new Gtk.ToolButton (null, null);
            list_toolbar_remove.icon_name = "list-remove-symbolic";
            list_toolbar_remove.clicked.connect (() => {
                list_box.remove(list_box.get_selected_row ());
            });
            list_toolbar.insert (list_toolbar_remove, -1);

            // SAVE Button
            var list_toolbar_save = new Gtk.ToolButton (null, null);
            list_toolbar_save.icon_name = "document-save-symbolic";
            list_toolbar_save.clicked.connect (() => {
                (list_box.get_selected_row () as Metronome.SettingRow).setting.beat = click.current_beat;
                (list_box.get_selected_row () as Metronome.SettingRow).setting.tempo = click.current_tempo;
            });
            list_toolbar.insert (list_toolbar_save, -1);

            sidebar.pack_start (list_scroll, true, true);
            sidebar.pack_start (list_toolbar, false, false);

            root.pack2 (sidebar, true, false);
        }

        private void do_click (click_interval interval, bool beat) {
            switch(interval) {
                case click_interval.semibreve:
                    led_green_right.visible = true;

                    if (beat) {
                        debug ("play beat");
                        sound_beat.set_state (State.NULL);
                        sound_beat.set ("volume", volume_semibreve.get_value ());
                        sound_beat.set ("uri", file_beat);
                        sound_beat.set_state (State.PLAYING);

                        led_red_left.visible = true;
                    } else {
                        if (volume_semibreve.get_value () > 0) {
                            debug ("play semibreve");
                            sound_semibreve.set_state (State.NULL);
                            sound_semibreve.set ("volume", volume_semibreve.get_value ());
                            sound_semibreve.set ("uri", file_click_1);
                            sound_semibreve.set_state (State.PLAYING);
                        }
                    }

                    GLib.Timeout.add (60, () => {
                        led_green_right.visible = false;
                        led_red_left.visible = false;
                        return false;
                    }, GLib.Priority.HIGH);
                    break;
                case click_interval.minim:
                    if (volume_minim.get_value () > 0) {
                        debug ("play minim");
                        sound_minim.set_state (State.NULL);
                        sound_minim.set ("volume", volume_minim.get_value ());
                        sound_minim.set ("uri", file_click_2);
                        sound_minim.set_state (State.PLAYING);
                    }
                    break;
                case click_interval.crotchet:
                    if (volume_crotchet.get_value () > 0) {
                        debug ("play crotchet");
                        sound_crotchet.set_state (State.NULL);
                        sound_crotchet.set ("volume", volume_crotchet.get_value ());
                        sound_crotchet.set ("uri", file_click_4);
                        sound_crotchet.set_state (State.PLAYING);
                    }
                    break;
                case click_interval.quaver:
                    if (volume_quaver.get_value () > 0) {
                        debug ("play quaver");
                        sound_quaver.set_state (State.NULL);
                        sound_quaver.set ("volume", volume_quaver.get_value ());
                        sound_quaver.set ("uri", file_click_8);
                        sound_quaver.set_state (State.PLAYING);
                    }
                    break;
            }
        }

        private bool set_new_tempo (uint new_tempo) {
            if (click.is_tempo_valid (new_tempo)) {
                if (click.current_tempo != new_tempo) {
                    click.current_tempo = new_tempo;

                    if (click.current_state == click_state.running) {
                        click.stop ();
                        click.start.begin ();
                    }
                }
                return true;
            }
            return false;
        }

        private bool set_new_beat (uint new_beat) {
            if (click.is_beat_valid (new_beat)) {
                if (click.current_beat != new_beat) {
                    click.current_beat = new_beat;

                    if (click.current_state == click_state.running) {
                        click.stop ();
                        click.start.begin ();
                    }
                }
                return true;
            }
            return false;
        }

        private void stop_sound () {
            sound_semibreve.set_state (State.NULL);
            sound_minim.set_state (State.NULL);
            sound_crotchet.set_state (State.NULL);
            sound_quaver.set_state (State.NULL);
            sound_beat.set_state (State.NULL);
        }

        private void load_settings () {
            tempo.value = settings.tempo;
            beat.value = settings.beat;
            volume_semibreve.set_value (settings.volume_semibreve);
            volume_minim.set_value (settings.volume_minim);
            volume_crotchet.set_value (settings.volume_crotchet);
            volume_quaver.set_value (settings.volume_quaver);

            foreach (string item in settings.custom_settings) {
                var custom_setting = Metronome.Setting.parse (item);
                if (custom_setting != null)
                    add_custom_setting (custom_setting);
            }
        }

        private void save_settings () {
            settings.tempo = (int) tempo.value;
            settings.beat = (int) beat.value;
            settings.volume_semibreve = volume_semibreve.get_value ();
            settings.volume_minim = volume_minim.get_value ();
            settings.volume_crotchet = volume_crotchet.get_value ();
            settings.volume_quaver = volume_quaver.get_value ();

            GLib.Array<string> custom_settings = new GLib.Array<string> ();
            foreach (Gtk.Widget row in list_box.get_children ()) {
                if (row is Metronome.SettingRow) {
                    debug ("save custom setting %s", (row as Metronome.SettingRow).setting.get_setting_string ());
                    custom_settings.append_val((row as Metronome.SettingRow).setting.get_setting_string ());
                }
            }
            settings.custom_settings = custom_settings.data;
        }

        private void add_custom_setting (Metronome.Setting custom_setting) {
            var row = new Metronome.SettingRow (custom_setting);
            list_box.add(row);
            row.get_style_context ().remove_class("button");
            debug ("add custom setting %s", custom_setting.title);
        }
    }
}
    public static int main (string [] args) {
        Gst.init (ref args);
        Gtk.init (ref args);
        var app = Metronome.MetronomeApp.instance;
        return app.run (args);
    }
